#!/usr/bin/env python3
# @Time    : 17-9-9 18:34
# @Author  : Wavky Huang
# @Contact : master@wavky.com
# @File    : mhcalendar.py

"""Calendar to manage your man-hour.

Usage:
    mhcalendar (-h | --help)
    mhcalendar (-V | --version)
    mhcalendar (-J | --Job) <required_manhour> <daily_work_hours> <hourly_pay> <max_daily_overhours>
    mhcalendar (-j | --job)
    mhcalendar (-M | --Month) <year> <month>
    mhcalendar (-m | --month)
    mhcalendar [-v | --verbose] [-C | --calendar] [--pre <precision>]
    mhcalendar (-c | --checkin) [<hours>]
    mhcalendar (-p | --pointer)
    mhcalendar --dayoff [--] [<date> ...]


Options:
    -h, --help        Show this screen.
    -V, --version     Show version.
    -J, --Job         Create or Update Job's file.
    -j, --job         Show your Job's file.
    -M, --Month       Create or Update Month data, default to current month if unset.
    -m, --month       Show the month which is processing.
    -v, --verbose     Show verbose message, it's about schedule processing now.
    -C, --calendar    Update and Show the calendar with man-hour schedule under specific precision.
    --pre             Specify a minimum time unit to schedule the man-hour. Default to 1(hour).
    -c, --checkin     Check in your man-hours into the date which is pointing to.
    -p, --pointer     Show the date which is pointing to.
                      It will shift to next day after you check in the man-hour.
    --dayoff          Schedule the dates to day off or reverse.


Parameter:
    Job
        required_manhour      Monthly manhour required by company.
        daily_work_hours      Daily work hours required by company.
        hourly_pay            Hourly pay offers by company.
        max_daily_overhours   How many hours you can work overtime per day, while minus means unlimited.

    checkin
        hours                 Hours to check in the date which is pointing to.
                              See --pointer in Options.
                              Default to the scheduled hours of that date if no specify.

    dayoff
        date...               Dates which is planing to take a day off, while minus means
                              that day switches to on duty.
                              You can only set dayoff the date which hasn't checkin.
                              Default to the date which is pointing to if no specify.
                              See --pointer in Options.
                              eg. --dayoff -- -9 -13 23 24


Simple Workflow:
    $ python3 mhcalendar.py -J ...      Create your Job file first.
                                        It will also update the holidays schedule at the first time.
    $ python3 mhcalendar.py             Check out the man-hour schedule of current month.
    $ python3 mhcalendar.py -c 8        Check in your man-hour of first day.
    $ python3 mhcalendar.py             Check out your schedule again to confirm your next day's work hours.

    $ python3 mhcalendar.py --dayoff -- 9 10 11 -13
                                        Schedule your day off or on duty days anytime before it comes.


"""
from docopt import docopt

import meta
import mhcalendar.log as log
import mhcalendar.time_elements as te
from mhcalendar import io
from mhcalendar.job import Job


def main():
    arguments = docopt(__doc__, version=meta.VERSION)
    io.prepare()
    log.VERBOSE = arguments['--verbose']
    if arguments['--Job']:
        job = Job(float(arguments['<required_manhour>']), float(arguments['<daily_work_hours>']),
                  float(arguments['<hourly_pay>']), float(arguments['<max_daily_overhours>']))
        schedule = io.Cache.restore_schedule()
        schedule.job = job
        io.Cache.cache_schedule(schedule)
        return
    if arguments['--job']:
        schedule = io.Cache.restore_schedule()
        check_schedule(schedule)
        print(schedule.job)
        return
    if arguments['--Month']:
        schedule = io.Cache.restore_schedule()
        month = te.Month(int(arguments['<year>']), int(arguments['<month>']))
        schedule.month = month
        io.Cache.cache_schedule(schedule)
        return
    if arguments['--month']:
        schedule = io.Cache.restore_schedule()
        print('Month: ', schedule.month.index['year'], '.', schedule.month.index['month'], sep='')
        return
    if arguments['--checkin']:
        schedule = io.Cache.restore_schedule()
        check_schedule(schedule)
        manhour = arguments['<hours>']
        day_to_checkin = schedule.month.next_day
        if not day_to_checkin:
            raise Exception()
        if manhour:
            day_to_checkin.checkin(float(manhour))
        else:
            day_to_checkin.checkin()
        io.Cache.cache_schedule(schedule)
        print("Date: {0}.{1}.{2} {3} \t Check in hours: {4}".format(day_to_checkin.date.year, day_to_checkin.date.month,
                                                                    day_to_checkin.date.day,
                                                                    day_to_checkin.date.strftime('%A'),
                                                                    day_to_checkin.checkin_manhour))
        return
    if arguments['--pointer']:
        schedule = io.Cache.restore_schedule()
        check_schedule(schedule)
        print(schedule.month.next_day)
        return
    if arguments['--dayoff']:
        schedule = io.Cache.restore_schedule()
        check_schedule(schedule)
        dates = arguments['<date>']

        if not dates or len(dates) == 0:
            dates = [-schedule.month.next_day.date.day]

        schedule.adjust(day_off=dates)
        io.Cache.cache_schedule(schedule)
        return

    # default to show calendar
    schedule_precision = arguments['<precision>'] or 1
    schedule = io.Cache.restore_schedule()
    check_schedule(schedule)

    schedule.schedule(schedule_precision)
    io.MHCalendarDrawer().draw(schedule)
    io.Cache.cache_schedule(schedule)


def check_schedule(schedule):
    if not schedule:
        raise Exception("Initialize error. Try again.")
    if not schedule.job:
        raise Exception("You haven't create your Job file yet.")
    if not schedule.month.next_day:
        raise Exception("This month's schedule is completed already.\nUpdate your Month data for next schedule.")


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# @Time    : 17-9-10 14:46
# @Author  : Wavky Huang
# @Contact : master@wavky.com
# @File    : log.py

"""
Log with switch by VERBOSE
"""
VERBOSE = False


def log(*values, sep=' ', end: str = '\n'):
    if VERBOSE:
        print(*values, sep=sep, end=end)



#!/usr/bin/env python3
# @Time    : 17-9-8 23:48
# @Author  : Wavky Huang
# @Contact : master@wavky.com
# @File    : io.py

"""
Process input and output
"""
import calendar
import json
import os
import pickle
from datetime import date
from urllib import request

from mhcalendar.time_elements import Schedule, Holiday, Month, dec_float

CONFIG_DIR = os.path.join(os.path.expanduser('~'), '.mhcalendar')


def exist(path):
    return os.path.exists(path)


def prepare():
    """
    Create config folder if not exist, cache holidays if not exist or out of date, and initialize schedule cache.
    """

    _check_config_path()

    # TODO: Maybe we can prepare holiday data for specific year
    if not _check_holiday_cache():
        holidays = update_holiday_schedule() or []
        Cache.cache_holidays(holidays)

    _init_schedule_cache()


def _check_config_path():
    """
    Create config folder if not exist
    """
    if not exist(CONFIG_DIR):
        os.makedirs(CONFIG_DIR)


def _check_holiday_cache():
    holiday_cache = Cache.restore_holidays()
    if holiday_cache and len(holiday_cache) > 0:
        thisyear = date.today().year
        if holiday_cache[0].year == str(thisyear):
            return True
    return False


def _init_schedule_cache():
    schedule_cache = Cache.restore_schedule()
    if not schedule_cache:
        today = date.today()
        month = Month(today.year, today.month)
        schedule = Schedule(None, month)
        Cache.cache_schedule(schedule)


class Cache:
    HOLIDAY_CACHE_NAME = 'holiday.cache'
    SCHEDULE_CACHE_NAME = 'schedule.cache'

    @classmethod
    def cache_holidays(cls, holidays):
        _check_config_path()
        path = os.path.join(CONFIG_DIR, Cache.HOLIDAY_CACHE_NAME)
        try:
            with open(path, 'w') as file:
                json.dump(holidays, file)
        except:
            print("Failure to open cache file:", path)

    @classmethod
    def cache_schedule(cls, schedule):
        _check_config_path()
        path = os.path.join(CONFIG_DIR, Cache.SCHEDULE_CACHE_NAME)
        try:
            with open(path, 'wb') as file:
                pickle.dump(schedule, file)
        except:
            print("Failure to open cache file:", path)

    @classmethod
    def restore_holidays(cls):
        """
        :return: Holiday list or None if no cache is found
        """
        path = os.path.join(CONFIG_DIR, Cache.HOLIDAY_CACHE_NAME)
        if exist(path):
            try:
                with open(path, 'r') as file:
                    jslist = json.load(file)
                    if jslist and isinstance(jslist, list):
                        holidays = []
                        for item in jslist:
                            holiday = Holiday(*item)
                            holidays.append(holiday)
                        return holidays
                    else:
                        return None
            except:
                print("Failure to open cache file:", path)
        else:
            return None

    @classmethod
    def restore_schedule(cls):
        """
        :return: Schedule object or None if no cache is found
        """
        path = os.path.join(CONFIG_DIR, Cache.SCHEDULE_CACHE_NAME)
        if exist(path):
            try:
                with open(path, 'rb') as file:
                    schedule = pickle.load(file)
                    return schedule if isinstance(schedule, Schedule) else None
            except:
                print("Failure to open cache file:", path)
        else:
            return None


class MHCalendarDrawer:
    """
    Output a monthly calendar.
    """

    def __init__(self, width=14):
        """

        :param width: minimum width is limited to 12, and MUST BE set as EVEN
        """
        self.width = 12 if width < 12 else width
        self.hr_line = '-' * (width + 1) * 7 + '-'
        self.seperator = '|' + ('-' * width + '|') * 7
        self.blank_line = str(' ' * ((width + 1) * 7 + 1))

    def __separate_line(self, week: str, end=''):
        week = list(week)
        distance_to_end = self.width * 7 + 8 - len(week)
        week += list(' ' * distance_to_end)
        week[::self.width + 1] = list('|' * 8)
        return ''.join(week) + end

    def __pipe(self, start, width, *texts, seperator='|', end=''):
        """
        The real name is Pile In piPE :P

        :param start: offset to start
        :param width: pipe's width
        :param texts: text to fill into a pipe
        :param seperator: seperator before and after, not count in pipe's width
        :return: a bamboo...
        """
        elements = [' ' * start, seperator]
        for t in texts:
            lent = len(t)
            if lent < width:
                t += ' ' * (width - lent)
            t += seperator
            elements.append(t)
        elements.append(end)
        return ''.join(elements)

    def __packup_week_schedule(self, week):
        first_weekday = week[0].date.weekday()
        start_index = first_weekday * (self.width + 1)
        pipes = list()

        holidays = ['* Holiday *'.center(self.width)
                    if week[i].holiday is not None else ' ' * self.width for i in range(len(week))]
        schedule_hours = ['Sched: {}'.format(week[i].scheduled_work_hours)
                          if not week[i].is_dayoff else '-' for i in range(len(week))]
        overtime_hours = ['OT: {}'.format(week[i].overtime)
                          if not week[i].is_dayoff else '-' for i in range(len(week))]
        checkin_hours = ['Checkin: {}'.format(week[i].checkin_manhour)
                         if not week[i].is_dayoff else '-' for i in range(len(week))]
        dayoff = ['Dayoff: {}'.format('Yes' if week[i].is_dayoff else 'No') for i in range(len(week))]
        done = ['Done: {}'.format('Yes' if week[i].is_past else 'No') for i in range(len(week))]
        pipes.append(self.__separate_line(self.__pipe(start_index, self.width, *holidays), '\n'))
        pipes.append(self.__separate_line(self.__pipe(start_index, self.width, *schedule_hours), '\n'))
        pipes.append(self.__separate_line(self.__pipe(start_index, self.width, *overtime_hours), '\n'))
        pipes.append(self.__separate_line(self.__pipe(start_index, self.width, *checkin_hours), '\n'))
        pipes.append(self.__separate_line(self.__pipe(start_index, self.width, *dayoff), '\n'))
        pipes.append(self.__separate_line(self.__pipe(start_index, self.width, *done), '\n'))
        return pipes

    def __decorate_today(self, month, week_line):
        if month.today:
            day = date.today().day
            index = week_line.find(' ' + str(day) + ' ')
            if index > 0:
                week_line = list(week_line)
                week_line[index - 1] = '['
                r_offset = 3 if day < 10 else 4
                week_line[index + r_offset] = ']'
                week_line = ''.join(week_line)
        return week_line

    def __print_holiday(self, schedule):
        holidays = schedule.month.holidays
        if holidays and len(holidays):
            for holiday in holidays:
                print('{year}.{month}.{day}  {name}'
                      .format(year=holiday.year, month=holiday.month, day=holiday.day, name=holiday.name))

    def __print_today(self, schedule):
        day = schedule.month.today
        day_name = date.today().strftime('%A')
        if day:
            holiday = '** {} **'.format(day.holiday.name) if day.holiday else ''
            checkin_or_dayoff = '\t Checkin: {}'.format(day.checkin_manhour) if not day.is_dayoff else '\t Day off'
            print('Today:', str(day.date), day_name, holiday, '\t Schedule(OT):', day.scheduled_work_hours,
                  '({})'.format(day.overtime), checkin_or_dayoff)
        else:
            print('today:', date.today(), day_name)

    def __print_manhour_expect(self, schedule):
        workdays = len(schedule.month.days) - len(schedule.dayoff_list)
        salary = schedule.job.required_manhour * schedule.job.hourly_pay
        print('Expecting:', 'Manhour/Workdays = {0}/{1}d'.format(schedule.job.required_manhour, workdays),
              '\t Salary = {}'.format(salary))

    def __print_manhour_fornow(self, schedule):
        print('For now:  ', 'Checkin manhour = {}'.format(schedule.checkin_manhour),
              '\t Remaining manhour = {}'.format(schedule.manhour_remain),
              '\t Overtime = {}'.format(schedule.overhours),
              '\t Salary = {}'.format(schedule.checkin_manhour * dec_float(schedule.job.hourly_pay)))

    def __print_manhour_absence(self, schedule):
        if schedule.manhour_absence > 0:
            print('These manhour can not be scheduled on this month:', schedule.manhour_absence)

    def draw(self, schedule: Schedule):
        cal = str(calendar.month(schedule.month.index['year'], schedule.month.index['month'], self.width))
        cal_lines = [' ' + w for w in cal.splitlines() if w != '']

        title = cal_lines.pop(0)
        day_of_week = cal_lines.pop(0)
        cal_weeks = cal_lines

        print('', title, self.hr_line, day_of_week, self.__separate_line(self.hr_line), sep='\n')
        for i in range(len(cal_weeks)):
            print(self.__separate_line(self.__decorate_today(schedule.month, cal_weeks[i])))
            print(*self.__packup_week_schedule(schedule.month.weeks[i]), sep='', end='')
            print(self.__separate_line(self.hr_line))
        print('(Sched = Schedule, OT = Overtime)')
        self.__print_holiday(schedule)
        print('')
        self.__print_today(schedule)
        self.__print_manhour_expect(schedule)
        self.__print_manhour_fornow(schedule)
        self.__print_manhour_absence(schedule)


def update_holiday_schedule():
    """
    request new schedule list of holidays this year.

    :return: new list of Holiday or None for update failure.
    """

    url = "http://calendar-service.net/cal?start_year={year}&start_mon=1&end_year={year}&end_mon=12\
&year_style=normal&month_style=numeric&wday_style=en&format=csv&holiday_only=1".format(year=date.today().year)
    print('Accessing network to request holiday data...')
    print('url: ' + url)

    try:
        with request.urlopen(url) as f:
            content = [line.decode('EUC-JP').replace('\n', '') for line in f.readlines()]
            del content[0]
            content = [line.split(',') for line in content]
            holidays = [Holiday(*line) for line in content]
            print('Success.')
            return holidays
    except:
        print("Holiday schedule request failure.")
        return None



#!/usr/bin/env python3
# @Time    : 17-9-2 01:53
# @Author  : Wavky Huang
# @Contact : master@wavky.com
# @File    : job.py

"""
Process information of the job.
"""


class Job:
    def __init__(self, required_manhour=0, daily_work_hours=0, hourly_pay=0, max_daily_overhours=0):
        """
        Define your job's condition.

        :param required_manhour: monthly manhour required by company
        :param daily_work_hours: daily work hours required by company
        :param hourly_pay: hourly pay offers by company
        :param max_daily_overhours: how many hours you can work overtime per day, while minus means unlimited
        """
        self.required_manhour = required_manhour
        self.daily_work_hours = daily_work_hours
        self.hourly_pay = hourly_pay
        self.max_daily_overhours = max_daily_overhours

        if max_daily_overhours < 0:
            self.max_daily_overhours = 24 - daily_work_hours

        if daily_work_hours + max_daily_overhours > 24:
            self.max_daily_overhours = 24 - daily_work_hours
            print("daily_work_hours + max_daily_overhours > 24, max_daily_overhours has been set to {0}.".format(
                self.max_daily_overhours))

    def __str__(self):
        return "Current Job: \t Require manhour = {0} \t Daily work hours = {1} \n\
\t\t Hourly pay = {2} \t\t Max daily overhours = {3}".format(self.required_manhour, self.daily_work_hours,
                                                             self.hourly_pay, self.max_daily_overhours)

    def __repr__(self):
        return self.__str__()



#!/usr/bin/env python3
# @Time    : 17-9-2 14:47
# @Author  : Wavky Huang
# @Contact : master@wavky.com
# @File    : time_elements.py

"""
Process overall information of one month.
"""
import calendar
from collections import namedtuple
from datetime import date, datetime, timezone, timedelta
from decimal import Decimal
from functools import reduce

from mhcalendar.job import Job
from mhcalendar.log import log

Holiday = namedtuple('Holiday',
                     ('year', 'month', 'day', 'year_name', 'year_count', 'weekday', 'weekday_number', 'name'))


class Month:
    def __init__(self, year, month):
        self.index = {'year': year, 'month': month}
        self.dates = list(
            filter(lambda date_: date_.month == month, calendar.Calendar().itermonthdates(year, month)))

        from mhcalendar.io import Cache
        holidays_one_year = Cache.restore_holidays() or []
        self.holidays = list(filter(lambda ho: ho.month == str(self.index['month']), holidays_one_year))
        self.days = self.__dates2days()
        self.weeks = self.__days2weeks()

    @property
    def today(self):
        """
        :return: the Day object of today, or None if today is not in this month
        """
        td = date.today()
        if td in self.dates:
            index = self.dates.index(td)
            return self.days[index]

    @property
    def next_day(self):
        """
        Obtain the next day that is not past.

        :return: object Day or None
        """
        for day in self.days:
            if not day.is_past:
                return day

    def __dates2days(self):
        """
        adapt dates to days

        :return: list of Day
        """

        days = []
        holidays = self.holidays.copy()
        for date_ in self.dates:
            holiday = None
            if len(holidays) > 0 and holidays[0].day == str(date_.day):
                holiday = holidays[0]
                del holidays[0]

            # weekday 5 means Saturday
            dayoff = holiday is not None or date_.weekday() >= 5
            days.append(Day(date_, holiday, dayoff))
        return days

    def __days2weeks(self):
        today = date.today()
        date_matrix = calendar.Calendar().monthdayscalendar(today.year, today.month)
        date_list = list(filter(lambda d: d > 0, calendar.Calendar().itermonthdays(today.year, today.month)))
        weeks = []
        for w in date_matrix:
            week = []
            weeks.append(week)
            for d in w:
                if d > 0:
                    week.append(self.days[date_list.index(d)])
        return weeks

    def __str__(self):
        return "MonthlyCalendar({year}, {month} days: {days})".format(year=self.index['year'],
                                                                      month=self.index['month'],
                                                                      days=self.days)

    def __repr__(self):
        return self.__str__()


class Day:
    def __init__(self, date_: date, holiday: Holiday = None, dayoff=False,
                 schedule=0, checkin=0, overtime=0, past=False):
        self.date = date_
        self.holiday = holiday
        self.is_dayoff = dayoff
        self.scheduled_work_hours = schedule
        self.checkin_manhour = checkin
        # overtime at past day stands for OT indeed, otherwise it only stands for scheduled-OT
        self.overtime = overtime
        self.is_past = past

    def checkin(self, hours=0, past=True):
        """
        Check in the man hours of this day.

        :param hours: man hours today, default 0 means check in as scheduled manhour
        :param past: if work today has done or not
        """
        self.checkin_manhour = hours if hours > 0 else self.scheduled_work_hours
        if past:
            self.overtime -= self.scheduled_work_hours - self.checkin_manhour
        self.is_past = past

    def dayoff(self):
        """
        You should only set day off the day which is NOT past.
        """
        self.is_dayoff = True
        self.checkin_manhour = 0
        self.schedule(0)
        self.overtime = 0

    def schedule(self, hours=0):
        """
        You should only schedule a day that is NOT past.

        :param hours:
        """
        self.scheduled_work_hours = hours

    def __str__(self):
        return "Day({date}, holiday={holiday}, dayoff={dayoff}, " \
               "scheduled={scheduled}, checkin={checkin}, past={past})" \
            .format(date=self.date, holiday=self.holiday, dayoff=self.is_dayoff,
                    scheduled=self.scheduled_work_hours, checkin=self.checkin_manhour, past=self.is_past)

    def __repr__(self):
        return self.__str__()


class Schedule:
    def __init__(self, job: [Job, None], month: Month):
        self.job = job
        self.month = month
        # how many hours you have worked this month (Decimal type or 0)
        self.checkin_manhour = 0
        # how many hours last you need to work this month (Decimal type or 0)
        self.manhour_remain = 0
        # how many overhours you have done for now (Decimal type or 0)
        self.overhours = 0
        # hours that CAN NOT be scheduled on this month
        self.manhour_absence = 0

    @property
    def dayoff_list(self):
        return list(filter(lambda day: day.is_dayoff, self.month.days))

    def adjust(self, *, day_off: [int] = ()):
        """
        Adjust the schedule by setting dates of dayoff.
        You can ONLY adjust the days that is not past.
        You need to re-schedule it before draw().

        :param day_off: dates which is planing to take a day off, while minus means that day switches to on duty
        """
        if len(day_off) == 0:
            return

        dayoff = sorted(set(filter(lambda d: d > 0, day_off)))
        onduty = sorted(set(abs(x) for x in filter(lambda d: d < 0, day_off)))

        invalid_dates = sorted(set(filter(lambda d: abs(d) > len(self.month.days) or d == 0, day_off)))
        if len(invalid_dates) > 0:
            print("Parameter out of range of month: {0}".format(invalid_dates))
            return

        conflict_dates = sorted(set(filter(lambda d: d in onduty, dayoff)))
        conflict_dates = [(d, -d) for d in conflict_dates]
        if len(conflict_dates) > 0:
            print("Parameter conflicts at: {0}".format(conflict_dates))
            return

        dayoff_days = {date: self.month.days[date - 1] for date in dayoff}
        onduty_days = {date: self.month.days[date - 1] for date in onduty}

        past_day = dict(filter(lambda item: item[1].is_past, dayoff_days.items()))
        past_day.update(dict(filter(lambda item: item[1].is_past, onduty_days.items())))
        if len(past_day) > 0:
            print("Illegal past dates: {0}".format(list(past_day.keys())))
            return

        dayoff_days = dayoff_days.values()
        onduty_days = onduty_days.values()

        for day in dayoff_days:
            day.dayoff()
        for day in onduty_days:
            day.is_dayoff = False

    @classmethod
    def __sort_day(cls, days):
        return sorted(days, key=lambda day: day.date)

    def __calculate_manhour_remain(self):
        days_past = set(filter(lambda day: day.is_past, self.month.days))
        days_remain = set(self.month.days) - days_past
        workdays_remain = set(filter(lambda day: not day.is_dayoff, days_remain))
        dayoff_remain = days_remain - workdays_remain

        if len(days_past) > 0:
            self.checkin_manhour = reduce(lambda a, b: a + b, [dec_float(day.checkin_manhour) for day in days_past])
            self.overhours = reduce(lambda a, b: a + b, [dec_float(day.overtime) for day in days_past])
        else:
            self.checkin_manhour = 0
            self.overhours = 0
        delta_manhour_remain = dec_float(self.job.required_manhour) - self.checkin_manhour
        self.manhour_remain = delta_manhour_remain if delta_manhour_remain > 0 else 0
        return self.__sort_day(workdays_remain), self.__sort_day(dayoff_remain), self.manhour_remain

    def __ceil_workhour_by_precision(self, workhour: Decimal, precision: Decimal):
        """
        make schedule time "ceil" to precision, only if it's not beyond the max overtime

        :param workhour:
        :param precision:
        :return: Decimal type
        """
        if precision > 0:
            if workhour % precision:
                pre_schedule_hours = (workhour // precision + 1) * precision
            else:
                pre_schedule_hours = workhour
        else:
            pre_schedule_hours = workhour

        max_daily_work_hours = dec_float(self.job.daily_work_hours) + dec_float(self.job.max_daily_overhours)
        pre_schedule_hours = max_daily_work_hours \
            if pre_schedule_hours > max_daily_work_hours else pre_schedule_hours

        return pre_schedule_hours

    def schedule(self, precision=1):
        """
        Update schedule time of workdays remaining.

        :param precision: minimum time unit to calculate the schedule time, default to 1 hour
        (eg. 0.25 means 15 mins). 0 means no limit.
        :return:
        """
        workdays_remain, dayoff_remain, manhour_remain = self.__calculate_manhour_remain()
        workdays_count = len(workdays_remain)

        log('workdays remaining:', workdays_count)
        for day in workdays_remain:

            log('date:', day.date, "\t manhour_remain:", manhour_remain)
            dec_daily_work_hours = dec_float(self.job.daily_work_hours)
            if manhour_remain > dec_daily_work_hours:
                avg_manhour_remain = Decimal(manhour_remain / workdays_count).quantize(Decimal('1.00'))
                workdays_count -= 1
                if avg_manhour_remain > dec_daily_work_hours:
                    schedule_hours = self.__ceil_workhour_by_precision(avg_manhour_remain, dec_float(precision))
                else:
                    schedule_hours = dec_daily_work_hours
                log('daily_avg_manhour_remain:', avg_manhour_remain)
            else:
                schedule_hours = dec_daily_work_hours

            day.schedule(float(schedule_hours))
            if schedule_hours > dec_daily_work_hours:
                day.overtime = float(schedule_hours - dec_daily_work_hours)
            else:
                day.overtime = 0

            manhour_remain -= schedule_hours

            log("schedule_hours:", schedule_hours, "\t manhour_remain:", manhour_remain, '\n')

        for day in dayoff_remain:
            day.schedule(0)
            day.overtime = 0

        self.manhour_absence = float(manhour_remain) if manhour_remain > 0 else 0


def timezone_date(tz=+9, area='Tokyo'):
    return datetime.now(tz=timezone(timedelta(hours=tz), area)).date()


def any(func, iterable):
    for x in iterable:
        if func(x):
            return True
    return False


def dec_float(number: float):
    return Decimal(str(number))



#!/usr/bin/env python3
# @Time    : 17-9-2 01:39
# @Author  : Wavky Huang
# @Contact : master@wavky.com
# @File    : example.py

"""
Example to run the program.
"""

import mhcalendar.io as io
import mhcalendar.time_elements as mh
from mhcalendar.job import Job

month = mh.Month(2017, 9)

job = Job(required_manhour=120, daily_work_hours=7.5, hourly_pay=2000, max_daily_overhours=2)
schedule = mh.Schedule(job, month)
schedule.schedule(0.5)

mhc = io.MHCalendarDrawer(width=14)
mhc.draw(schedule)

month.next_day.checkin(7.5)
month.next_day.checkin()
month.next_day.checkin()
month.next_day.checkin()

schedule.adjust(day_off=[6, 12, -9, -17])
schedule.schedule(0.5)

mhc.draw(schedule)



#!/usr/bin/env python3
# @Time    : 17-9-9 18:27
# @Author  : Wavky Huang
# @Contact : master@wavky.com
# @File    : setup.py.py

"""

"""
from setuptools import setup, find_packages

from meta import *

setup(
    name=PROJECT_NAME,
    version=VERSION,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    description='Calendar to manage your man-hour.',
    long_description=open('README.rst').read(),
    url='https://github.com/wavky/ManHourCalendar',
    license='MIT',
    keywords=['man-hour', 'manhour', 'man hour', 'calendar', 'schedule'],
    python_requires='>=3',
    packages=find_packages(),
    py_modules=['meta'],
    install_requires=['docopt>=0.6.2', 'pypandoc>=1.4'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Topic :: Office/Business :: Scheduling',
        'Topic :: Utilities',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Environment :: Console'
    ],
    entry_points={
        'console_scripts': [
            'mhcalendar = scripts.mhcalendar:main'
        ]
    }
)



#!/usr/bin/env python3
# @Time    : 17-9-10 01:08
# @Author  : Wavky Huang
# @Contact : master@wavky.com
# @File    : meta.py

"""

"""
PROJECT_NAME = 'ManHourCalendar'
VERSION = '0.9.1b2'
AUTHOR = 'Wavky Huang'
AUTHOR_EMAIL = 'master@wavky.com'
